@@ -50,10 +50,15 @@
 #include "treemerge.h"
 #include "utilities.h"
 
-#if HAVE_BTRFS_H
-#include <linux/btrfs.h>
-#include <sys/ioctl.h>
-#endif
+/* define paranoia levels */
+static const RmDigestType RM_PARANOIA_LEVELS[] = {RM_DIGEST_METRO,
+                                                  RM_DIGEST_METRO256,
+                                                  RM_DIGEST_HIGHWAY256,
+                                                  RM_DEFAULT_DIGEST,
+                                                  RM_DIGEST_PARANOID,
+                                                  RM_DIGEST_PARANOID};
+static const int RM_PARANOIA_NORMAL = 3;  /*  must be index of RM_DEFAULT_DIGEST */
+static const int RM_PARANOIA_MAX = 5;
 
 static void rm_cmd_show_version(void) {
     fprintf(stderr, "version %s compiled: %s at [%s] \"%s\" (rev %s)\n", RM_VERSION,
@@ -233,105 +238,6 @@ static int rm_cmd_maybe_switch_to_hasher(int argc, const char **argv) {
     return EXIT_SUCCESS;
 }
 
-static void rm_cmd_btrfs_clone_usage(void) {
-    rm_log_error(_("Usage: rmlint --btrfs-clone [-r] source dest\n"));
-}
-
-static void rm_cmd_btrfs_clone(const char *source, const char *dest,
-                               const gboolean read_only) {
-#if HAVE_BTRFS_H
-    struct {
-        struct btrfs_ioctl_same_args args;
-        struct btrfs_ioctl_same_extent_info info;
-    } extent_same;
-    memset(&extent_same, 0, sizeof(extent_same));
-
-    int source_fd = rm_sys_open(source, O_RDONLY);
-    if(source_fd < 0) {
-        rm_log_error_line(_("btrfs clone: failed to open source file"));
-        return;
-    }
-
-    extent_same.info.fd = rm_sys_open(dest, read_only ? O_RDONLY : O_RDWR);
-    if(extent_same.info.fd < 0) {
-        rm_log_error_line(_("btrfs clone: error %i: failed to open dest file.%s"),
-                          errno,
-                          read_only ? "" : _("\n\t(if target is a read-only snapshot "
-                                             "then -r option is required)"));
-        rm_sys_close(source_fd);
-        return;
-    }
-
-    struct stat source_stat;
-    fstat(source_fd, &source_stat);
-
-    guint64 bytes_deduped = 0;
-    gint64 bytes_remaining = source_stat.st_size;
-    int ret = 0;
-    while(bytes_deduped < (guint64)source_stat.st_size && ret == 0 &&
-          extent_same.info.status == 0 && bytes_remaining) {
-        extent_same.args.dest_count = 1;
-        extent_same.args.logical_offset = bytes_deduped;
-        extent_same.info.logical_offset = bytes_deduped;
-
-        /* BTRFS_IOC_FILE_EXTENT_SAME has an internal limit at 16MB */
-        extent_same.args.length = MIN(16 * 1024 * 1024, bytes_remaining);
-        if(extent_same.args.length == 0) {
-            extent_same.args.length = bytes_remaining;
-        }
-
-        ret = ioctl(source_fd, BTRFS_IOC_FILE_EXTENT_SAME, &extent_same);
-        if(ret == 0 && extent_same.info.status == 0) {
-            bytes_deduped += extent_same.info.bytes_deduped;
-            bytes_remaining -= extent_same.info.bytes_deduped;
-        }
-    }
-
-    rm_sys_close(source_fd);
-    rm_sys_close(extent_same.info.fd);
-
-    if(ret < 0) {
-        ret = errno;
-        rm_log_error_line(_("BTRFS_IOC_FILE_EXTENT_SAME returned error: (%d) %s"), ret,
-                          strerror(ret));
-    } else if(extent_same.info.status == -22 && read_only && getuid()) {
-        rm_log_error_line(_("Need to run as root user to clone to a read-only snapshot"));
-    } else if(extent_same.info.status < 0) {
-        rm_log_error_line(_("BTRFS_IOC_FILE_EXTENT_SAME returned status %d for file %s"),
-                          extent_same.info.status, dest);
-    } else if(bytes_remaining > 0) {
-        rm_log_info_line(_("Files don't match - not cloned"));
-    }
-#else
-    (void)source;
-    (void)dest;
-    (void)read_only;
-    rm_log_error_line(_("rmlint was not compiled with btrfs support."))
-
-#endif
-}
-
-static int rm_cmd_maybe_btrfs_clone(RmSession *session, int argc, const char **argv) {
-    if(argc > 0 && g_strcmp0("--btrfs-clone", argv[1]) == 0) {
-        /* treat as a btrfs clone subcommand... */
-        if(!rm_session_check_kernel_version(session, 4, 2)) {
-            rm_log_warning_line("This needs at least linux >= 4.2.");
-        } else if(argc == 5 && g_strcmp0("-r", argv[2]) == 0) {
-            /* -r option for deduping read-only snapshots */
-            /* TODO: add check for root user permissions */
-            rm_cmd_btrfs_clone(argv[3], argv[4], TRUE);
-        } else if(argc == 4) {
-            rm_cmd_btrfs_clone(argv[2], argv[3], FALSE);
-        } else {
-            /* malformed command */
-            rm_cmd_btrfs_clone_usage();
-        }
-        /* return EXIT_FAILURE to indicate not to go ahead with main rmlint call */
-        return EXIT_FAILURE;
-    }
-    return EXIT_SUCCESS;
-}
-
 /* clang-format off */
 static const struct FormatSpec {
     const char *id;
@@ -452,25 +358,36 @@ static GLogLevelFlags VERBOSITY_TO_LOG_LEVEL[] = {[0] = G_LOG_LEVEL_CRITICAL,
                                                   [3] = G_LOG_LEVEL_MESSAGE |
                                                         G_LOG_LEVEL_INFO,
                                                   [4] = G_LOG_LEVEL_DEBUG};
+static bool rm_cmd_read_paths_from_stdin(RmSession *session, bool is_prefd,
+                                         bool null_separated) {
+    char delim = null_separated ? 0 : '\n';
+
+    size_t buf_len = PATH_MAX;
+    char *path_buf = malloc(buf_len * sizeof(char));
 
-static bool rm_cmd_read_paths_from_stdin(RmSession *session, bool is_prefd) {
-    char path_buf[PATH_MAX];
-    char *tokbuf = NULL;
     bool all_paths_read = true;
 
+    int path_len;
+
     /* Still read all paths on errors, so the user knows all paths that failed */
-    while(fgets(path_buf, PATH_MAX, stdin)) {
-        all_paths_read &=
-            rm_cfg_add_path(session->cfg, is_prefd, strtok_r(path_buf, "\n", &tokbuf));
+    while((path_len = getdelim(&path_buf, &buf_len, delim, stdin)) >= 0) {
+        if(path_len > 0) {
+            /* replace returned delimiter with null */
+            if (path_buf[path_len - 1] == delim) {
+                path_buf[path_len - 1] = 0;
+            }
+            all_paths_read &= rm_cfg_add_path(session->cfg, is_prefd, path_buf);
+        }
     }
 
+    free(path_buf);
     return all_paths_read;
 }
 
 static bool rm_cmd_parse_output_pair(RmSession *session, const char *pair,
                                      GError **error) {
-    rm_assert_gentle(session);
-    rm_assert_gentle(pair);
+    g_assert(session);
+    g_assert(pair);
 
     char *separator = strchr(pair, ':');
     char *full_path = NULL;
@@ -820,19 +737,26 @@ static gboolean rm_cmd_parse_timestamp_file(const char *option_name,
         memset(stamp_buf, 0, sizeof(stamp_buf));
 
         if(fgets(stamp_buf, sizeof(stamp_buf), stamp_file) != NULL) {
-            success = rm_cmd_parse_timestamp(option_name, g_strstrip(stamp_buf), session,
-                                             error);
+            success = rm_cmd_parse_timestamp(
+                option_name,
+                g_strstrip(stamp_buf),
+                session,
+                error
+            );
+
+            if(!success) {
+                return false;
+            }
+
             plain = rm_cmd_timestamp_is_plain(stamp_buf);
         }
 
         fclose(stamp_file);
     } else {
-        /* Cannot read... */
+        /* Cannot read a stamp file, assume we gonna creae it. */
         plain = false;
-    }
-
-    if(!success) {
-        return false;
+        success = true;
+        rm_log_info_line(_("No stamp file at `%s`, will create one after this run."), timestamp_path);
     }
 
     rm_fmt_add(session->formats, "stamp", timestamp_path);
@@ -855,32 +779,17 @@ static void rm_cmd_set_verbosity_from_cnt(RmCfg *cfg, int verbosity_counter) {
 static void rm_cmd_set_paranoia_from_cnt(RmCfg *cfg, int paranoia_counter,
                                          GError **error) {
     /* Handle the paranoia option */
-    switch(paranoia_counter) {
-    case -2:
-        cfg->checksum_type = RM_DIGEST_XXHASH;
-        break;
-    case -1:
-        cfg->checksum_type = RM_DIGEST_BASTARD;
-        break;
-    case 0:
-        /* leave users choice of -a (default) */
-        break;
-    case 1:
-#if HAVE_SHA512
-        cfg->checksum_type = RM_DIGEST_SHA512;
-#else
-        cfg->checksum_type = RM_DIGEST_SHA256;
-#endif
-        break;
-    case 2:
-        cfg->checksum_type = RM_DIGEST_PARANOID;
-        break;
-    default:
+    int index = paranoia_counter + RM_PARANOIA_NORMAL;
+
+    if(index < 0 || index > RM_PARANOIA_MAX) {
         if(error && *error == NULL) {
             g_set_error(error, RM_ERROR_QUARK, 0,
-                        _("Only up to -pp or down to -PP flags allowed"));
+                        _("Only up to -%.*s or down to -%.*s flags allowed"),
+                        RM_PARANOIA_MAX - RM_PARANOIA_NORMAL, "ppppp", RM_PARANOIA_NORMAL,
+                        "PPPPP");
         }
-        break;
+    } else {
+        cfg->checksum_type = RM_PARANOIA_LEVELS[index];
     }
 }
 
@@ -903,9 +812,6 @@ static gboolean rm_cmd_parse_algorithm(_UNUSED const char *option_name,
     if(cfg->checksum_type == RM_DIGEST_UNKNOWN) {
         g_set_error(error, RM_ERROR_QUARK, 0, _("Unknown hash algorithm: '%s'"), value);
         return false;
-    } else if(cfg->checksum_type == RM_DIGEST_BASTARD) {
-        session->hash_seed1 = time(NULL) * (GPOINTER_TO_UINT(session));
-        session->hash_seed2 = GPOINTER_TO_UINT(&session);
     }
     return true;
 }
@@ -944,6 +850,12 @@ static gboolean rm_cmd_parse_limit_mem(_UNUSED const char *option_name,
     return (rm_cmd_parse_mem(size_spec, error, &session->cfg->total_mem));
 }
 
+static gboolean rm_cmd_parse_read_buf_len(_UNUSED const char *option_name,
+                                          const gchar *size_spec, RmSession *session,
+                                          GError **error) {
+    return (rm_cmd_parse_mem(size_spec, error, &session->cfg->read_buf_len));
+}
+
 static gboolean rm_cmd_parse_sweep_size(_UNUSED const char *option_name,
                                         const gchar *size_spec, RmSession *session,
                                         GError **error) {
@@ -971,7 +883,8 @@ static gboolean rm_cmd_parse_clamp_top(_UNUSED const char *option_name, const gc
 static gboolean rm_cmd_parse_progress(_UNUSED const char *option_name,
                                       _UNUSED const gchar *value, RmSession *session,
                                       _UNUSED GError **error) {
-    rm_fmt_clear(session->formats);
+    rm_fmt_remove_by_name(session->formats, "pretty");
+
     rm_fmt_add(session->formats, "progressbar", "stdout");
     rm_fmt_add(session->formats, "summary", "stdout");
 
@@ -1183,10 +1096,11 @@ static gboolean rm_cmd_parse_rankby(_UNUSED const char *option_name,
 }
 
 static gboolean rm_cmd_parse_replay(_UNUSED const char *option_name,
-                                    _UNUSED const gchar *x, RmSession *session,
+                                    const gchar *json_path, RmSession *session,
                                     _UNUSED GError **error) {
     session->cfg->replay = true;
     session->cfg->cache_file_structs = true;
+    rm_cfg_add_path(session->cfg, false, json_path);
     return true;
 }
 
@@ -1210,6 +1124,22 @@ static gboolean rm_cmd_parse_equal(_UNUSED const char *option_name,
     return true;
 }
 
+static gboolean rm_cmd_parse_btrfs_clone(_UNUSED const char *option_name,
+                                   _UNUSED const gchar *x, RmSession *session,
+                                   _UNUSED GError **error) {
+    rm_log_warning_line("option --btrfs-clone is deprecated, use --dedupe");
+    session->cfg->dedupe = true;
+    return true;
+}
+
+static gboolean rm_cmd_parse_btrfs_readonly(_UNUSED const char *option_name,
+                                   _UNUSED const gchar *x, RmSession *session,
+                                   _UNUSED GError **error) {
+    rm_log_warning_line("option --btrfs-readonly is deprecated, use --dedupe-readonly");
+    session->cfg->dedupe_readonly = true;
+    return true;
+}
+
 static bool rm_cmd_set_cwd(RmCfg *cfg) {
     /* Get current directory */
     char cwd_buf[PATH_MAX + 1];
@@ -1240,14 +1170,16 @@ static bool rm_cmd_set_cmdline(RmCfg *cfg, int argc, char **argv) {
 static bool rm_cmd_set_paths(RmSession *session, char **paths) {
     bool is_prefd = false;
     bool all_paths_valid = true;
+    bool stdin_paths_preferred = false;
 
     RmCfg *cfg = session->cfg;
 
     /* Check the directory to be valid */
     for(int i = 0; paths && paths[i]; ++i) {
         if(strcmp(paths[i], "-") == 0) {
-            /* option '-' means read paths from stdin */
-            all_paths_valid &= rm_cmd_read_paths_from_stdin(session, is_prefd);
+            cfg->read_stdin = true;
+            /* remember whether to treat stdin paths as preferred paths */
+            stdin_paths_preferred = is_prefd;
         } else if(strcmp(paths[i], "//") == 0) {
             /* the '//' separator separates non-preferred paths from preferred */
             is_prefd = !is_prefd;
@@ -1258,7 +1190,13 @@ static bool rm_cmd_set_paths(RmSession *session, char **paths) {
 
     g_strfreev(paths);
 
-    if(cfg->path_count == 0 && all_paths_valid) {
+    if(cfg->read_stdin || cfg->read_stdin0) {
+        /* option '-' means read paths from stdin */
+        all_paths_valid &=
+            rm_cmd_read_paths_from_stdin(session, stdin_paths_preferred, cfg->read_stdin0);
+    }
+
+    if(g_slist_length(cfg->paths) == 0 && all_paths_valid) {
         /* Still no path set? - use `pwd` */
         rm_cfg_add_path(session->cfg, is_prefd, cfg->iwd);
     }
@@ -1272,8 +1210,7 @@ static bool rm_cmd_set_outputs(RmSession *session, GError **error) {
         g_set_error(error, RM_ERROR_QUARK, 0,
                     _("Specifiyng both -o and -O is not allowed"));
         return false;
-    } else if(session->output_cnt[0] < 0 && session->output_cnt[1] < 0 &&
-              !rm_fmt_len(session->formats)) {
+    } else if(session->output_cnt[0] < 0 && session->cfg->progress_enabled == false) {
         rm_cmd_set_default_outputs(session);
     }
 
@@ -1302,7 +1239,6 @@ static char * rm_cmd_find_own_executable_path(RmSession *session, char **argv) {
 /* Parse the commandline and set arguments in 'settings' (glob. var accordingly) */
 bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
     RmCfg *cfg = session->cfg;
-    gboolean clone = FALSE;
 
     /* Handle --gui before all other processing,
      * since we need to pass other args to the python interpreter.
@@ -1317,10 +1253,6 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         return false;
     }
 
-    if(rm_cmd_maybe_btrfs_clone(session, argc, (const char **)argv) == EXIT_FAILURE) {
-        return false;
-    }
-
     /* List of paths we got passed (or NULL) */
     char **paths = NULL;
 
@@ -1352,11 +1284,11 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         {"newer-than"       , 'N' , 0        , G_OPTION_ARG_CALLBACK , FUNC(timestamp)      , _("Newer than timestamp")                 , "STAMP"}               ,
         {"config"           , 'c' , 0        , G_OPTION_ARG_CALLBACK , FUNC(config)         , _("Configure a formatter")                , "FMT:K[=V]"}           ,
 
-        /* Non-trvial switches */
+        /* Non-trivial switches */
         {"progress" , 'g' , EMPTY , G_OPTION_ARG_CALLBACK , FUNC(progress) , _("Enable progressbar")                   , NULL} ,
         {"loud"     , 'v' , EMPTY , G_OPTION_ARG_CALLBACK , FUNC(loud)     , _("Be more verbose (-vvv for much more)") , NULL} ,
         {"quiet"    , 'V' , EMPTY , G_OPTION_ARG_CALLBACK , FUNC(quiet)    , _("Be less verbose (-VVV for much less)") , NULL} ,
-        {"replay"   , 'Y' , EMPTY , G_OPTION_ARG_CALLBACK , FUNC(replay)   , _("Re-output a json file")                , "path/to/rmlint.json"} ,
+        {"replay"   , 'Y' , 0     , G_OPTION_ARG_CALLBACK , FUNC(replay)   , _("Re-output a json file")                , "path/to/rmlint.json"} ,
         {"equal"    ,  0 ,  EMPTY , G_OPTION_ARG_CALLBACK , FUNC(equal)    , _("Test for equality of PATHS")           , "PATHS"}           ,
 
         /* Trivial boolean options */
@@ -1377,16 +1309,23 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         {"honour-dir-layout"        , 'j'  , EMPTY     , G_OPTION_ARG_CALLBACK  , FUNC(honour_dir_layout)        , _("Only find directories with same file layout")                          , NULL}     ,
         {"perms"                    , 'z'  , OPTIONAL  , G_OPTION_ARG_CALLBACK  , FUNC(permissions)              , _("Only use files with certain permissions")                              , "[RWX]+"} ,
         {"no-hardlinked"            , 'L'  , DISABLE   , G_OPTION_ARG_NONE      , &cfg->find_hardlinked_dupes    , _("Ignore hardlink twins")                                                , NULL}     ,
+        {"keep-hardlinked"          , 0    , 0         , G_OPTION_ARG_NONE      , &cfg->keep_hardlinked_dupes    , _("Keep hardlink that are linked to any original")                        , NULL}     ,
         {"partial-hidden"           , 0    , EMPTY     , G_OPTION_ARG_CALLBACK  , FUNC(partial_hidden)           , _("Find hidden files in duplicate folders only")                          , NULL}     ,
         {"mtime-window"             , 'Z'  , 0         , G_OPTION_ARG_DOUBLE    , &cfg->mtime_window             , _("Consider duplicates only equal when mtime differs at max. T seconds")  , "T"}      ,
+        {"stdin0"                   , '0'  , 0         , G_OPTION_ARG_NONE      , &cfg->read_stdin0              , _("Read null-separated file list from stdin")                             , NULL}     ,
+        {"no-backup"                , 0    , 0         , G_OPTION_ARG_NONE      , &cfg->no_backup                , _("Do not create backups of previous result files")                       , NULL}     ,
+
+        /* COW filesystem deduplication support */
+        {"dedupe"                   , 0    , 0         , G_OPTION_ARG_NONE      , &cfg->dedupe                   , _("Dedupe matching extents from source to dest (if filesystem supports)") , NULL}     ,
+        {"dedupe-readonly"          , 'r'  , 0         , G_OPTION_ARG_NONE      , &cfg->dedupe_readonly          , _("(--dedupe option) even dedupe read-only snapshots (needs root)")       , NULL}     ,
+        {"is-reflink"               , 0    , 0         , G_OPTION_ARG_NONE      , &cfg->is_reflink               , _("Test if two files are reflinks (share same data extents)")             , NULL}     ,
 
         /* Callback */
         {"show-man" , 'H' , EMPTY , G_OPTION_ARG_CALLBACK , rm_cmd_show_manpage , _("Show the manpage")            , NULL} ,
         {"version"  , 0   , EMPTY , G_OPTION_ARG_CALLBACK , rm_cmd_show_version , _("Show the version & features") , NULL} ,
         /* Dummy option for --help output only: */
         {"gui"         , 0 , 0 , G_OPTION_ARG_NONE , NULL   , _("If installed, start the optional gui with all following args")                 , NULL},
-        {"hash"        , 0 , 0 , G_OPTION_ARG_NONE , NULL   , _("Work like sha1sum for all supported hash algorithms (see also --hash --help)") , NULL}                                            ,
-        {"btrfs-clone" , 0 , 0 , G_OPTION_ARG_NONE , &clone , _("Clone extents from source to dest, if extents match")                          , NULL} ,
+        {"hash"        , 0 , 0 , G_OPTION_ARG_NONE , NULL   , _("Work like sha1sum for all supported hash algorithms (see also --hash --help)") , NULL},
 
         /* Special case: accumulate leftover args (paths) in &paths */
         {G_OPTION_REMAINING , 0 , 0 , G_OPTION_ARG_FILENAME_ARRAY , &paths , ""   , NULL}   ,
@@ -1415,8 +1354,9 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         {"clamp-low"              , 'q' , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(clamp_low)              , "Limit lower reading barrier"                                 , "P"}    ,
         {"clamp-top"              , 'Q' , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(clamp_top)              , "Limit upper reading barrier"                                 , "P"}    ,
         {"limit-mem"              , 'u' , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(limit_mem)              , "Specify max. memory usage target"                            , "S"}    ,
-        {"sweep-size"             , 'u' , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(sweep_size)             , "Specify max. bytes per pass when scanning disks"             , "S"}    ,
-        {"sweep-files"            , 'u' , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(sweep_count)            , "Specify max. file count per pass when scanning disks"        , "S"}    ,
+        {"read-buffer-len"        , 0   , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(read_buf_len)           , "Specify read buffer length in bytes"                         , "S"}    ,
+        {"sweep-size"             , 0   , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(sweep_size)             , "Specify max. bytes per pass when scanning disks"             , "S"}    ,
+        {"sweep-files"            , 0   , HIDDEN           , G_OPTION_ARG_CALLBACK , FUNC(sweep_count)            , "Specify max. file count per pass when scanning disks"        , "S"}    ,
         {"threads"                , 't' , HIDDEN           , G_OPTION_ARG_INT64    , &cfg->threads                , "Specify max. number of hasher threads"                       , "N"}    ,
         {"threads-per-disk"       , 0   , HIDDEN           , G_OPTION_ARG_INT      , &cfg->threads_per_disk       , "Specify number of reader threads per physical disk"          , NULL}   ,
         {"write-unfinished"       , 'U' , HIDDEN           , G_OPTION_ARG_NONE     , &cfg->write_unfinished       , "Output unfinished checksums"                                 , NULL}   ,
@@ -1432,10 +1372,17 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         {"fake-abort"             , 0   , HIDDEN           , G_OPTION_ARG_NONE     , &cfg->fake_abort             , "Simulate interrupt after 10% shredder progress"              , NULL}   ,
         {"buffered-read"          , 0   , HIDDEN           , G_OPTION_ARG_NONE     , &cfg->use_buffered_read      , "Default to buffered reading calls (fread) during reading."   , NULL}   ,
         {"shred-never-wait"       , 0   , HIDDEN           , G_OPTION_ARG_NONE     , &cfg->shred_never_wait       , "Never waits for file increment to finish hashing"            , NULL}   ,
+        {"no-sse"                 , 0   , HIDDEN           , G_OPTION_ARG_NONE     , &cfg->no_sse                 , "Don't use SSE accelerations"                                 , NULL}   ,
         {"no-mount-table"         , 0   , DISABLE | HIDDEN , G_OPTION_ARG_NONE     , &cfg->list_mounts            , "Do not try to optimize by listing mounted volumes"           , NULL}   ,
         {NULL                     , 0   , HIDDEN           , 0                     , NULL                         , NULL                                                          , NULL}
     };
 
+    const GOptionEntry deprecated_option_entries[] = {
+        {"btrfs-clone"              , 0    , EMPTY | HIDDEN      , G_OPTION_ARG_CALLBACK      , FUNC(btrfs_clone)         , "Deprecated, use --dedupe instead"                  , NULL},
+        {"btrfs-readonly"           , 0    , EMPTY | HIDDEN      , G_OPTION_ARG_CALLBACK      , FUNC(btrfs_readonly)      , "Deprecated, use --dedupe-readonly instead"         , NULL},
+        {NULL                       , 0    , HIDDEN              , 0                          , NULL                      , NULL                                                , NULL}
+    };
+
     /* clang-format on */
 
     /* Initialize default verbosity */
@@ -1443,17 +1390,17 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
 
     if(!rm_cmd_set_cwd(cfg)) {
         g_set_error(&error, RM_ERROR_QUARK, 0, _("Cannot set current working directory"));
-        goto failure;
+        goto cleanup;
     }
 
     if(!rm_cmd_set_cmdline(cfg, argc, argv)) {
         g_set_error(&error, RM_ERROR_QUARK, 0, _("Cannot join commandline"));
-        goto failure;
+        goto cleanup;
     }
 
     /* Attempt to find out path to own executable.
      * This is used in the shell script to call the executable
-     * for special modes like --btrfs-clone or --equal.
+     * for special modes like --dedupe or --equal.
      * We want to make sure the installed version has this
      * */
     cfg->full_argv0_path = rm_cmd_find_own_executable_path(session, argv);
@@ -1462,6 +1409,18 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
     // OPTION PARSING //
     ////////////////////
 
+    /* TODO: move subcommands to separate option parser
+     * e.g.
+     * Usage:
+     * rmlint [options] <paths>...
+     * rmlint --subcommand [options]
+     * Subcommands (must be first arg):
+     *      --dedupe      Dedupe matching extents from source to dest (if filesystem supports)
+     *      --is-reflink  Test if two files are reflinks
+     *      --gui         Launch rmlint gui
+     * For help on subcommands use rmlint --<subcommand> --help
+     *
+     */
     option_parser = g_option_context_new(
         _("[TARGET_DIR_OR_FILES …] [//] [TAGGED_TARGET_DIR_OR_FILES …] [-]"));
     g_option_context_set_translation_domain(option_parser, RM_GETTEXT_PACKAGE);
@@ -1472,13 +1431,17 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         "inversed", "inverted", "Options that enable defaults", session, NULL);
     GOptionGroup *unusual_group =
         g_option_group_new("unusual", "unusual", "Unusual options", session, NULL);
+    GOptionGroup *deprecated_group =
+        g_option_group_new("deprecated", "deprecated", "Deprecated options", session, NULL);
 
     g_option_group_add_entries(main_group, main_option_entries);
     g_option_group_add_entries(main_group, inversed_option_entries);
     g_option_group_add_entries(main_group, unusual_option_entries);
+    g_option_group_add_entries(deprecated_group, deprecated_option_entries);
 
     g_option_context_add_group(option_parser, inversion_group);
     g_option_context_add_group(option_parser, unusual_group);
+    g_option_context_add_group(option_parser, deprecated_group);
     g_option_context_set_main_group(option_parser, main_group);
     g_option_context_set_summary(option_parser,
                                  _("rmlint finds space waste and other broken things on "
@@ -1497,13 +1460,24 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
     g_option_group_set_error_hook(main_group, (GOptionErrorFunc)rm_cmd_on_error);
 
     if(!g_option_context_parse(option_parser, &argc, &argv, &error)) {
-        goto failure;
+        goto cleanup;
     }
 
-    if(clone) {
-        /* should not get here */
-        rm_cmd_btrfs_clone_usage();
-        session->cmdline_parse_error = TRUE;
+    if(!rm_cmd_set_paths(session, paths)) {
+        error = g_error_new(RM_ERROR_QUARK, 0, _("Not all given paths are valid. Aborting"));
+        goto cleanup;
+    }
+
+    if(cfg->replay && (cfg->dedupe || cfg->is_reflink)) {
+        error = g_error_new(
+            RM_ERROR_QUARK, 0,
+            _("--replay (-Y) is incompatible with --dedupe or --is-reflink")
+        ); goto cleanup;
+    }
+
+    if(cfg->dedupe) {
+        /* dedupe session; regular rmlint configs are ignored */
+        goto cleanup;
     }
 
     /* Silent fixes of invalid numeric input */
@@ -1540,6 +1514,8 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
         cfg->with_stdout_color = cfg->with_stderr_color = 0;
     }
 
+    g_assert(!(cfg->follow_symlinks && cfg->see_symlinks));
+
     if(cfg->keep_all_tagged && cfg->keep_all_untagged) {
         error = g_error_new(
             RM_ERROR_QUARK, 0,
@@ -1547,16 +1523,33 @@ bool rm_cmd_parse_args(int argc, char **argv, RmSession *session) {
     } else if(cfg->skip_start_factor >= cfg->skip_end_factor) {
         error = g_error_new(RM_ERROR_QUARK, 0,
                             _("-q (--clamp-low) should be lower than -Q (--clamp-top)"));
-    } else if(!rm_cmd_set_paths(session, paths)) {
-        error = g_error_new(RM_ERROR_QUARK, 0, _("Not all given paths are valid. Aborting"));
     } else if(!rm_cmd_set_outputs(session, &error)) {
         /* Something wrong with the outputs */
-    } else if(cfg->follow_symlinks && cfg->see_symlinks) {
-        rm_log_error("Program error: Cannot do both follow_symlinks and see_symlinks");
-        rm_assert_gentle_not_reached();
+    } else if(cfg->keep_all_tagged && cfg->must_match_untagged) {
+        error = \
+            g_error_new(
+                RM_ERROR_QUARK, 0,
+                _(
+                    "-k and -M should not be specified at the same time " \
+                    "(see also: https://github.com/sahib/rmlint/issues/244)"
+                )
+        );
+    } else if(cfg->keep_all_untagged && cfg->must_match_tagged) {
+        error = \
+            g_error_new(
+                RM_ERROR_QUARK, 0,
+                _(
+                    "-K and -m should not be specified at the same time " \
+                    "(see also: https://github.com/sahib/rmlint/issues/244)"
+                )
+        );
     }
 
-failure:
+#if HAVE_BUILTIN_CPU_SUPPORTS && HAVE_MM_CRC32_U64
+    rm_digest_enable_sse(!cfg->no_sse && __builtin_cpu_supports("sse4.2"));
+#endif
+
+cleanup:
     if(error != NULL) {
         rm_cmd_on_error(NULL, NULL, session, &error);
     }
@@ -1607,7 +1600,7 @@ int rm_cmd_main(RmSession *session) {
 
     rm_fmt_set_state(session->formats, RM_PROGRESS_STATE_INIT);
 
-    if(session->cfg->replay) {
+    if(cfg->replay) {
         return rm_cmd_replay_main(session);
     }
 
@@ -1618,7 +1611,7 @@ int rm_cmd_main(RmSession *session) {
     }
 
     if(session->mounts == NULL) {
-       rm_log_debug_line("No mount table created.");
+        rm_log_debug_line("No mount table created.");
     }
 
     session->mds = rm_mds_new(cfg->threads, session->mounts, cfg->fake_pathindex_as_disk);
@@ -1629,7 +1622,7 @@ int rm_cmd_main(RmSession *session) {
                       g_timer_elapsed(session->timer, NULL), session->total_files);
 
     if(cfg->merge_directories) {
-        rm_assert_gentle(cfg->cache_file_structs);
+        g_assert(cfg->cache_file_structs);
 
         /* Currently we cannot use -D and the cloning on btrfs, since this assumes the same layout
          * on two dupicate directories which is likely not a valid assumption.
@@ -1648,7 +1641,11 @@ int rm_cmd_main(RmSession *session) {
             return EXIT_FAILURE;
         }
 
-        session->dir_merger = rm_tm_new(session);
+        RmTreeMerger *t = session->dir_merger = rm_tm_new(session);
+        if(!t) {
+            rm_log_error_line(_("Failed to complete setup for merging directories"));
+            return EXIT_FAILURE;
+        }
     }
 
     if(session->total_files < 2 && session->cfg->run_equal_mode) {
@@ -1656,7 +1653,30 @@ int rm_cmd_main(RmSession *session) {
         return EXIT_FAILURE;
     }
 
+    /* some optimisations for rmlint --equal */
+    if(cfg->run_equal_mode && session->total_files == 2) {
+        /* check if the two files are hardlinks or reflinks or some such */
+        g_assert(cfg->paths);
+        RmPath *a = cfg->paths->data;
+        g_assert(cfg->paths->next);
+        RmPath *b = cfg->paths->next->data;
+        switch(rm_util_link_type(a->path, b->path)) {
+        case RM_LINK_HARDLINK:
+        case RM_LINK_REFLINK:
+        case RM_LINK_PATH_DOUBLE:
+        case RM_LINK_SAME_FILE:
+            session->equal_exit_code = EXIT_SUCCESS;
+            cfg->find_duplicates = FALSE;
+            cfg->merge_directories = FALSE;
+            rm_log_debug_line("got match via rm_offsets_match");
+            break;
+        default:
+            break;
+        }
+    }
+
     if(session->total_files >= 1) {
+
         rm_fmt_set_state(session->formats, RM_PROGRESS_STATE_PREPROCESS);
         rm_preprocess(session);
 
@@ -1694,7 +1714,7 @@ int rm_cmd_main(RmSession *session) {
         exit_state = EXIT_FAILURE;
     }
 
-    if(exit_state == EXIT_SUCCESS && cfg->run_equal_mode)  {
+    if(exit_state == EXIT_SUCCESS && cfg->run_equal_mode) {
         return session->equal_exit_code;
     }
 
