@@ -418,7 +418,7 @@ static int rm_pp_cmp_criterion(unsigned char criterion, const RmFile *a, const R
         return sign * cmp;
     }
     default:
-        rm_assert_gentle_not_reached();
+        g_assert_not_reached();
         return 0;
     }
 }
@@ -500,8 +500,8 @@ static gboolean rm_pp_check_path_double(RmFile *file, GHashTable *unique_paths_t
         g_hash_table_add(unique_paths_table, key);
         return FALSE;
     }
-    RmFile *match_double = match_double_key->file;
-    rm_assert_gentle(match_double != file);
+
+    g_assert(match_double_key->file != file);
 
     rm_path_double_free(key);
     rm_file_destroy(file);
@@ -569,7 +569,7 @@ static gboolean rm_pp_handle_inode_clusters(_UNUSED gpointer key, GQueue *inode_
 
         /* hardlink cluster are counted as filtered files since they are either
          * ignored or treated as automatic duplicates depending on settings (so
-         * no effort eaither way); rm_pp_handle_hardlink will either free or bundle
+         * no effort either way); rm_pp_handle_hardlink will either free or bundle
          * the hardlinks depending on value of headfile->hardlinks.is_head.
          */
         session->total_filtered_files -= rm_util_queue_foreach_remove(
@@ -579,7 +579,7 @@ static gboolean rm_pp_handle_inode_clusters(_UNUSED gpointer key, GQueue *inode_
     /* update counters */
     rm_fmt_set_state(session->formats, RM_PROGRESS_STATE_PREPROCESS);
 
-    rm_assert_gentle(inode_cluster->length <= 1);
+    g_assert(inode_cluster->length <= 1);
     if(inode_cluster->length == 1) {
         session->tables->size_groups->data = g_slist_prepend(
             session->tables->size_groups->data, inode_cluster->head->data);
@@ -608,8 +608,8 @@ static RmOff rm_pp_handler_other_lint(const RmSession *session) {
         for(GList *iter = list; iter; iter = iter->next) {
             RmFile *file = iter->data;
 
-            rm_assert_gentle(file);
-            rm_assert_gentle(type == file->lint_type);
+            g_assert(file);
+            g_assert(type == file->lint_type);
 
             num_handled++;
 
@@ -650,7 +650,7 @@ void rm_preprocess(RmSession *session) {
 
     /* split into file size groups; for each size, remove path doubles and bundle
      * hardlinks */
-    rm_assert_gentle(all_files->head);
+    g_assert(all_files->head);
     RmFile *file = g_queue_pop_head(all_files);
     RmFile *current_size_file = file;
     guint removed = 0;
