@@ -73,7 +73,7 @@ def test_paranoia(shell):
     create_link('a', 'hardlink_a', symlink=False)
 
     head, *data, footer = run_rmlint(
-        '-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
+        '-S a --no-backup -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
     )
 
     assert footer['duplicate_sets'] == 1
@@ -93,7 +93,7 @@ def test_paranoia(shell):
         handle.write('xxx')
 
     head, *data, footer = run_rmlint(
-        '-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
+        '-S a --no-backup -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
     )
 
     assert footer['duplicate_sets'] == 1
@@ -107,7 +107,7 @@ def test_paranoia(shell):
     os.remove(os.path.join(TESTDIR_NAME, 'a'))
 
     text = run_shell_script(shell, sh_script, '-d', '-p')
-    head, *data, footer = run_rmlint('-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))
+    head, *data, footer = run_rmlint('-S a --no-backup -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))
 
     assert 'original has disappeared' in text
 
@@ -141,7 +141,7 @@ def test_hardlink_duplicate_directories(shell):
 
     sh_path = os.path.join(TESTDIR_NAME, "result.sh")
     header, *data, footer = run_rmlint(
-        "-D -S a -c sh:link -o sh:{}".format(sh_path),
+        "-D -S a -c sh:hardlink -o sh:{}".format(sh_path),
     )
     assert len(data) == 2
     assert data[0]["path"].endswith("dir_a")
@@ -230,3 +230,32 @@ def test_remove_empty_dirs_with_dupe_dirs(shell, inverse_order):
         assert data[1]["is_original"] is False
 
     _check_if_empty_dirs_deleted(shell, inverse_order, sh_path, data)
+
+@with_setup(usual_setup_func, usual_teardown_func)
+@parameterized([("sh", ), ("bash", ), ("dash", )])
+def test_cleanup_emptydirs(shell):
+    create_file('xxx', 'dir1/a')
+
+    # create some ugly dir names
+    names = [ 'escape me [please?]', '上野洋子, 吉野裕司, 浅井裕子 & 河越重義', '天谷大輔', 'Аркона',
+            'let\'s nest',
+            'let\'s nest/a level',
+            'let\'s nest/a level/[or two]' ]
+    for dirname in names:
+        create_file('xxx', '{}/b'.format(dirname))
+
+    head, *data, footer = run_rmlint('-S a -T df -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))
+
+    assert footer['duplicate_sets'] == 1
+    assert footer['total_lint_size'] == 3 * len(names)
+    assert footer['total_files'] == 1 + len(names)
+    assert footer['duplicates'] == len(names)
+
+    # run rmlint.sh with -c option (should clean up empty dirs after deleting 'b' files).
+    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')
+    text = run_shell_script(shell, sh_path, "-dc")
+
+    assert os.path.exists(os.path.join(TESTDIR_NAME, 'dir1/a'))
+
+    for dirname in names:
+        assert (not os.path.exists(os.path.join(TESTDIR_NAME, dirname)))
